""" Version and year of the audioio package.
"""

__version__ = '1.0.0'
""" Current version of the audioio package. """

__year__ = '2024'
""" Year of the current audioio version. """

__pdoc__ = {}
__pdoc__['__version__'] = True
__pdoc__['__year__'] = True

